@@ -71,7 +71,18 @@
 # define GetModuleFileName GetModuleFileNameA
 #endif
 
-#if !(defined _WIN32 && !defined __CYGWIN__) && !(defined __EMX__)
+/* We have special code for two types of system: non-Cygwin Windows, and
+   Linux where dladdr is in a separate library (uClibc and glibc < 2.34).
+   Need glibc >= 2, for getline().
+
+   Otherwise, use dladdr.
+*/
+#if defined __linux__ && (defined __UCLIBC__ || ((__GLIBC__ >= 2) && (__GLIBC_MINOR__ < 34)))
+#define _GL_USE_PROCFS 1
+#elif (defined _WIN32 && !defined __CYGWIN__) || defined __EMX__
+#define _GL_USE_WIN32 1
+#elif _GL_DLADDR_IN_LIBC
+#define _GL_USE_DLADDR 1
 #include <dlfcn.h>
 #endif
 
@@ -322,7 +333,6 @@ static char *shared_library_fullname;
 
 #if defined _WIN32 && !defined __CYGWIN__
 /* Native Windows only.
-
    On Cygwin, it is better to use dladdr, than to use native Windows
    API and cygwin_conv_to_posix_path, because it supports longer file
    names (see <https://cygwin.com/ml/cygwin/2011-01/msg00410.html>).  */
@@ -406,21 +416,69 @@ _DLL_InitTerm (unsigned long hModule, unsigned long ulFlag)
 static void
 find_shared_library_fullname ()
 {
+#if _GL_USE_PROCFS
+  /* Linux has /proc/self/maps. glibc 2 and uClibc have the getline()
+     function.
+     But it is costly: ca. 0.3 ms.  */
+  FILE *fp;
+
+  /* Open the current process' maps file.  It describes one VMA per line.  */
+  fp = fopen ("/proc/self/maps", "r");
+  if (fp)
+    {
+      unsigned long address = (unsigned long) &find_shared_library_fullname;
+      for (;;)
+        {
+          unsigned long start, end;
+          int c;
+
+          if (fscanf (fp, "%lx-%lx", &start, &end) != 2)
+            break;
+          if (address >= start && address <= end - 1)
+            {
+              /* Found it.  Now see if this line contains a filename.  */
+              while (c = getc (fp), c != EOF && c != '\n' && c != '/')
+                continue;
+              if (c == '/')
+                {
+                  size_t size;
+                  int len;
+
+                  ungetc (c, fp);
+                  shared_library_fullname = NULL; size = 0;
+                  len = getline (&shared_library_fullname, &size, fp);
+                  if (len >= 0)
+                    {
+                      /* Success: filled shared_library_fullname.  */
+                      if (len > 0 && shared_library_fullname[len - 1] == '\n')
+                        shared_library_fullname[len - 1] = '\0';
+                    }
+                }
+              break;
+            }
+          while (c = getc (fp), c != EOF && c != '\n')
+            continue;
+        }
+      fclose (fp);
+    }
+#elif _GL_USE_DLADDR
   Dl_info info;
   dladdr (find_shared_library_fullname, &info);
   if (info.dli_fname != NULL)
       shared_library_fullname = strdup (info.dli_fname);
+#endif
 }
 
 #endif /* Native Windows / EMX / Unix */
 
 /* Return the full pathname of the current shared library.
    Return NULL if unknown.
-   Guaranteed to work only on Linux, EMX, Cygwin, and native Windows.  */
+   Guaranteed to work only on Linux, EMX, Cygwin, native Windows, and
+   systems with dladdr in libc.  */
 static char *
 get_shared_library_fullname ()
 {
-#if !(defined _WIN32 && !defined __CYGWIN__) && !defined __EMX__
+#if !_GL_USE_WIN32
   static bool tried_find_shared_library_fullname;
   if (!tried_find_shared_library_fullname)
     {
